unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Grids, DBGrids, DB, DBClient, StdCtrls, Provider;

type
  TForm1 = class(TForm)
    ds1: TDataSource;
    btTest: TButton;
    dtProvided: TClientDataSet;
    dtProvidedField_Data: TIntegerField;
    gr1: TDBGrid;
    prSource: TDataSetProvider;
    dtSource: TClientDataSet;
    dtSourceField_Data: TIntegerField;
    dtSourceField_InternalCalc: TIntegerField;
    Label1: TLabel;
    Label2: TLabel;
    DBGrid1: TDBGrid;
    ds2: TDataSource;
    Memo1: TMemo;
    Label3: TLabel;
    procedure FormCreate(Sender: TObject);
    procedure btTestClick(Sender: TObject);
  private
    dt1, dt2 :TClientDataSet;
    RecCount :Integer;
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

Uses
  DSIntf;

Const
  { Error codes.  This constants are based on native error codes used
    internally by MIDAS. }
  errBase_InvalidReq = $2700;
  errCode_OutOfRange = 1;
  dberr_OutOfRange = errBase_InvalidReq + errCode_OutOfRange;

Type
  TFixedClientDataSet = Class (TClientDataSet)
    Protected
      Function CheckStatus (Const Status, Value :DBResult) :Boolean;
      Function CreateCursor :IDSCursor;
      Function CursorRecNumber (Const Cursor :IDSCursor) :Cardinal;
      Function GetFieldDescs :TFieldDescList;
      Procedure SetBaseOrder (Const Cursor :IDSCursor);
      Procedure AddDataPacket (Const Data :OLEVariant; HitEOF :Boolean);
        Override;
  End;

  { TFixedClientDataSet implementation }

  Function TFixedClientDataSet.CheckStatus (Const Status, Value :DBResult)
    :Boolean;
  Begin
    If Status = Value Then
      Result := True
    Else
    Begin
      Check (Status);  // Exception if Status <> 0
      Result := False;  // Status = 0
    End;
  End;

  Function TFixedClientDataSet.CreateCursor :IDSCursor;
  Begin
    CreateDbClientObject (ClsID_DSCursor, IDSCursor, Result);
    Result.InitCursor (DSBase);
  End;

  Function TFixedClientDataSet.CursorRecNumber (Const Cursor :IDSCursor)
    :Cardinal;
  Begin
    Check (Cursor.GetRecordNumber (Result));
  End;

  // Based on TCustomClientDataSet.InternalInitFieldDefs method
  Function TFixedClientDataSet.GetFieldDescs :TFieldDescList;
  Var
    Props :DSProps;
  Begin
    DSBase.SetProp (dspropCompressArrays, 1);
    DSBase.GetProps (Props);
    SetLength (Result, Props.iFields);
    DSBase.GetFieldDescs (PDSFldDesc (Result));
  End;

  Procedure TFixedClientDataSet.SetBaseOrder (Const Cursor :IDSCursor);
  Const
    BaseOrder = '_BaseOrder';
  Var
    Index :DSIDXDesc;

    Function Activate :DBResult;
    Begin
      Result := Cursor.UseIndexOrder (BaseOrder);
    End;
  Begin
    If Not CheckStatus (Activate, dberr_NoSuchIndex) Then
      Exit;

    { We add the "natural" base index (rows as they were created/appended).
      NOTE: the default order (szDefault_Order) is not necessarily equal to
      this "base order". }
    ZeroMemory (@Index, SizeOf (Index));
    StrLCopy (Index.szName, BaseOrder, SizeOf (Index.szName) - 1);
    Check (DSBase.CreateIndex (Index));

    Check (Activate);
  End;

  Procedure TFixedClientDataSet.AddDataPacket (Const Data :OLEVariant;
    HitEOF :Boolean);
  Var
    AuxCursor :IDSCursor;
    FieldDescs :TFieldDescList;
    FirstRec, I, J :Integer;
  Begin
    { NOTE: The IDSBase.AppendData method does not copy internally
      calculated fields, but, if called multiple times, it sets their
      "blank" flag incorrectly (MIDAS bug).  We correct this problem, by
      activating that flag on the added records. }

    { We need a clean cursor (not filtered or ranged) in "base order",
      with the purpose of to get the internal record number of the first
      row added }
    AuxCursor := CreateCursor;
    SetBaseOrder (AuxCursor);
    Check (AuxCursor.MoveToEOF);

    // Current last record (or BOF if the cursor is empty)
    CheckStatus (AuxCursor.MoveRelative (-1), dberr_BOF);

    // We append the row packet to internal record list
    Inherited AddDataPacket (Data, HitEOF);

    // Go to first record added (if there is one)
    If Not CheckStatus (AuxCursor.MoveRelative (1), dberr_EOF) Then
    Begin
      FirstRec := CursorRecNumber (AuxCursor);
      FieldDescs := GetFieldDescs;  // Field descriptors

      For I := 0 To High (FieldDescs) Do
        If FieldDescs [I].bCalculated Then  // InternalCalc field
          { From the first record added up to the last one + 1 (the limit
            for the IDSBase.PutBlank method is the internal "iRecNoNext"
            index), we put the field to "blank" by simply activating its
            blank flag in each record }
          For J := FirstRec To MaxInt Do
            If CheckStatus (DSBase.PutBlank (Nil, J,
            FieldDescs [I].iFieldID, 1), dberr_OutOfRange) Then
              Break;
    End;
  End;

  Type
    TDataSetAccess = Class (TDataSet);
procedure TForm1.FormCreate(Sender: TObject);
  Function PrepareDataSet (Const DataSet :TClientDataSet) :TClientDataSet;
  Var
    I :Integer;
  Begin
    DataSet.FieldDefs.Assign (dtSource.FieldDefs);
    TDataSetAccess (DataSet).CreateFields;
    DataSet.SetProvider (prSource);
    DataSet.Open;

    For I := 301 To 305 Do
      DataSet.AppendRecord ([I, I]);

    Result := DataSet;
  End;
begin
  dtProvided.CreateDataSet;
  dtSource.Open;
  dt1 := TClientDataSet.Create (Self);
  ds1.DataSet := PrepareDataSet (dt1);
  dt2 := TFixedClientDataSet.Create (Self);
  ds2.DataSet := PrepareDataSet (dt2);
end;

procedure TForm1.btTestClick(Sender: TObject);
Var
  I :Integer;
begin
  dtSource.Close;
  dtSource.Open;

  For I := 1 To 8 Do
  Begin
    Inc (RecCount);
    dtSource.AppendRecord ([RecCount, RecCount]);
  End;

  dt1.AppendData (dtSource.Data, True);
  dt1.Last;
  dt2.AppendData (dtSource.Data, True);
  dt2.Last;
end;

end.
